-- Orbiter Mesh Load Script for gmax and 3DSMax v1.1 --
-- July 2003 by Alexander Blass (mindblast@gmx.de)   --
-------------------------------------------------------
-- Version History:
-- v1.0 
--    * Initial version 
-- v1.1 
--    * Improved import of texture coordinates, no unwrap_UVW modifiers are
--      created anymore and the texture coordinates are stored directly in 
--      the mesh instead.
--    * The bLoadTextureData Variable removed since loading with textures is 
--      now lightning fast. :)
--    * All Faces are now created with smoothing group 1 only, otherwise gmax
--      automatically assigns groups wich leads to vertices with multiple normal
--      vectors and thus to inconsistencies with the export script (files would
--      get larger and larger with each import/export cycle). 
-----------------------------------------------------------------------------

struct SMeshGroup (meshobj, iMat, iTex)
strFileName = getOpenFileName caption:"Load Orbiter Mesh File" types:"Orbiter Mesh File *.msh|*.msh"

if strFileName != undefined then 
(
	fsMeshFile = openFile strFileName mode:"r";
	if fsMeshFile == undefined then return();
	if readLine fsMeshFile != "MSHX1" then return(); -- make sure this is a real orbiter mesh file
	strLine = readLine fsMeshFile			-- read line from file... 
	strChopLine = filterString strLine " "  -- and divide it in substrings
	iNumGroups = strChopLine[2] as integer  -- read Number of Groups

	meshGroups = #()
	for iAktGroup = 1 to iNumGroups do
	(
		newGroup = SMeshGroup()
		iNoNormal = 0
		do
		(
			strLine = readLine fsMeshFile
			strChopLine = filterString strLine " \t"
			if strChopLine[1] == "MATERIAL" then			-- read Material Index
				newGroup.iMat = strChopLine[2] as integer + 1 
				-- + 1 because Arrays in gmax are 1 based, so Material 1 would be our "default material"

			if strChopLine[1] == "TEXTURE" then				-- read Texture Index
				newGroup.iTex = strChopLine[2] as integer
				-- this time not + 1 because TEXTURE 0 means no texture

			if strChopLine[1] == "NONORMAL" then		
				iNoNormal = 3
		)
		while strChopLine[1] != "GEOM" 
		
		-- according to the .msh documentation the material and texture is 
		-- inherited from the previous group, if not specified, first group must specify them
		if newGroup.iMat == undefined then newGroup.iMat = meshGroups[iAktGroup-1].iMat  
		if newGroup.iTex == undefined then newGroup.iTex = meshGroups[iAktGroup-1].iTex

		-- get number of vertices and faces from the GEOM tag
		iVerts = strChopLine[2] as integer
		iFaces = strChopLine[3] as integer

		-- if we have Group names after the GROUP tag separated with a colon (i.e. atlantis.msh), we name our mesh-objects after them
		if strChopLine[4] != undefined then 
		(
			strChopLine = filterString strLine ";"
			meshName = substring strChopLine[2] 2 -1
		)
		else -- otherwise we name it "Group <nr>"
		(
			meshName = "Group " + iAktGroup as string
		)
		print ("Reading Group " + iAktGroup as string + "/" + iNumGroups as string + " : " + meshName)
		arrTexCoords = #()
		arrVerts = #()
		arrNormals = #()
		arrFaces = #()

		for iAktVert = 1 to iVerts do
		(
			strLine = readLine fsMeshFile
			strChopLine = filterString strLine " "

			-- read vertices and normal vectors into the arrays and perform left hand to right hand conversion
			arrVerts[iAktVert] = [strChopLine[1] as float * -1.0, strChopLine[2] as float, strChopLine[3] as float]
			arrNormals[iAktVert] = [strChopLine[4] as float * -1.0, strChopLine[5] as float, strChopLine[6] as float]

			-- if the group has a texture, read texture coordinates
			if newGroup.iTex != undefined and newGroup.iTex > 0 then
				arrTexCoords[iAktVert] = [strChopLine[7-iNoNormal] as float, strChopLine[8-iNoNormal] as float * -1.0 + 1.0, 0]
		)
		for iAktFace = 1 to iFaces do 
		(
			strLine = readLine fsMeshFile
			strChopLine = filterString strLine " "

			-- read in faces
			arrFaces[iAktFace] = [strChopLine[1] as integer + 1, strChopLine[3] as integer + 1, strChopLine[2] as integer + 1]
		)
		-- create the mesh
		newMesh = mesh vertices:arrVerts faces:arrFaces tverts:arrTexCoords
		
		-- set the normal vectors
		for iCurVert = 1 to iVerts do
			setNormal newMesh iCurVert arrNormals[iCurVert]

		-- create texture faces if we have a texture
		if newGroup.iTex != 0 then
		(
			buildTVFaces newMesh false
			for iAktFace = 1 to arrFaces.count do
				setTVFace newMesh iAktFace arrFaces[iAktFace]
		)
		newMesh.name = meshName

		-- to make sure gmax doesn't calculate its own smoothing groups, we 
		-- set all the faces to smoothing group 1 
		for iAktFace = 1 to newMesh.numfaces do
			setFaceSmoothGroup newMesh iAktFace 1

		update newMesh

		newGroup.meshobj = newMesh

		-- store the mesh together with its material and texture indices in an array
		-- this is used later for mapping materials to the meshes
		meshGroups[iAktGroup] = newGroup
	)

	-- now create the material array
	-- index 1 is the standard material (white diffuse and opaque)
	newMat = standard()
	newMat.name = "default material"
	newMat.diffuse = color 255 255 255
	arrMaterials = #(newMat)

	strLine = readLine fsMeshFile
	strChopLine = filterString strLine " "
	if strChopLine[1] == "MATERIALS" then 
	(
		print "Reading Materials"
		-- get number of materials
		iNumMat = strChopLine[2] as integer
		-- read in the material names and create the materials 
		for iAktMat = 2 to iNumMat+1 do
		(
			arrMaterials[iAktMat] = standard()
			arrMaterials[iAktMat].name = readLine fsMeshFile
		)

		-- now read in the material specifications
		-- assuming that materials are in the same order as above
		for iAktMat = 2 to iNumMat+1 do
		(
			strLine = readLine fsMeshFile
			strChopLine = filterString strLine " "
			if strChopLine[1] == "MATERIAL" then
			(
												
				strLine = readLine fsMeshFile
				strChopLine = filterString strLine " "

				-- colors are converted to 255 range
				arrMaterials[iAktMat].diffuse = color (strChopLine[1] as float * 255 as integer) (strChopLine[2] as float * 255 as integer) (strChopLine[3] as float * 255 as integer)

				-- opacity is read from the diffuse color definition only and converted to percentage representation
				arrMaterials[iAktMat].opacity = strChopLine[4] as float * 100 as integer
				strLine = readLine fsMeshFile
				strChopLine = filterString strLine " "

				-- read ambient color
				arrMaterials[iAktMat].ambient = color (strChopLine[1] as float * 255 as integer) (strChopLine[2] as float * 255 as integer) (strChopLine[3] as float * 255 as integer)
				strLine = readLine fsMeshFile
				strChopLine = filterString strLine " "

				-- read specular color
				arrMaterials[iAktMat].specular = color (strChopLine[1] as float * 255 as integer) (strChopLine[2] as float * 255 as integer) (strChopLine[3] as float * 255 as integer)

				-- if specified read specular level
				if strChopLine[5] != undefined then 
					arrMaterials[iAktMat].specularlevel = strChopLine[5] as float
				strLine = readLine fsMeshFile
				strChopLine = filterString strLine " "

				-- read emissive color
				arrMaterials[iAktMat].useSelfIllumColor = on
				arrMaterials[iAktMat].selfIllumColor = color (strChopLine[1] as float * 255 as integer) (strChopLine[2] as float * 255 as integer) (strChopLine[3] as float * 255 as integer)
			)
		)
	)
	strLine = readLine fsMeshFile
	strChopLine = filterString strLine " "
	if strChopLine[1] == "TEXTURES" then 
	(
		print "Reading Textures"
		iNumTex = strChopLine[2] as integer
		strTexNames = #()

		-- read texture file names
		for iAktTex = 1 to iNumTex do
		(
			strLine = readLine fsMeshFile

			-- filenames are stored without extension (.dds) 
			strChopLine = filterString strLine "."
			strTexNames[iAktTex] = strChopLine[1]
		)
	)
	close fsMeshFile

	arrTexturedMaterials = #()
	arrMatTexKeys = #()
	for iAktGroup = 1 to iNumGroups do
	(
		aktGroup = meshGroups[iAktGroup]

		-- since gmax stores texture maps only as part of a material definition
		-- we have to create a gmax material for every material/texture combination
		-- that is actually used in the mesh file
		-- first we create a unique key for the combination
		iMatTexKey = (aktGroup.iMat) * 1000 + aktGroup.iTex

		-- then look it up in an array
		iArrIndex = findItem arrMatTexKeys iMatTexKey

		-- if the combination wasn't found, we create a new material with the texture
		if iArrIndex == 0 then 
		(
			append arrMatTexKeys iMatTexKey
			iArrIndex = arrMatTexKeys.count
			newTexturedMaterial = copy arrMaterials[aktGroup.iMat]
			if aktGroup.iTex > 0 then
			(
				-- if a texture is specified, we try loading a .bmp file
				-- with the texture name from the directory ..\textures relative to 
				-- the dir the mesh file is loaded from
				strMeshDir = getFilenamePath strFileName
				strTexDirChop = filterString strMeshDir "\\"
				strTexDir = ""
				for iDirIndex =1 to strTexDirChop.count-1 do
					strTexDir += strTexDirChop[iDirIndex] + "\\"
				strTexDir += "textures\\"

				-- unfortunately gmax doesn't support .dds files so the textures
				-- of a mesh have to be converted to .bmp format manually to be used in gmax
				newTexturedMaterial.diffuseMap = Bitmaptexture fileName:(strTexDir + strTexNames[aktGroup.iTex] + ".bmp")

				-- materials with a texture are marked with an additional comment in the material name 
				-- however if the string is already part of the material name (from a previous export)
				-- we don't append it a second time
				if (matchPattern newTexturedMaterial.name pattern:("*with Tex: " + strTexNames[aktGroup.iTex])) == false then
					newTexturedMaterial.name += (" with Tex: " + strTexNames[aktGroup.iTex])

				-- make sure the texturemap is visible in the viewport
				showTextureMap newTexturedMaterial true
			)

			-- store the new material in an array, the indices of which correspond
			-- to the arrMatTexKeys array
			arrTexturedMaterials[iArrIndex] = newTexturedMaterial
		)

		-- finally assign the material to the mesh
		meshGroups[iAktGroup].meshobj.material = arrTexturedMaterials[iArrIndex]		
	)
	print "Done"
)
