-- Orbiter Mesh Export Script for 3DSMax v1.11          --
-- July 2003 by Alexander Blass (mindblast@gmx.de)      --
----------------------------------------------------------
-- Version History:
-- v1.0 Initial version 
-- v1.1 
--    * All scene objects are now copied and the copies automatically converted
--      to Editable_Mesh, meaning the scene can contain any kind of objects
--      (non geometry objects i.e lighsources and the like will be skipped)
--    * No unwrap_UVW modifier is needed anymore to generate the texture
--      coordinates, these are taken directly from the object now.
--    * Multiple normal vectors on a vertex (because of smoothing groups) 
--      are now exported correctly to orbiter by splitting the vertex up. 
--    * Export scripts for gmax and 3DSMax are now separate with 3DSMax version
--      saving to a file and gmax version dumping the .msh file contents
--      to the Maxscript Listener Window
-- v1.11
--    * Bugfix: generation of texture mapping coordinates fixed 
-----------------------------------------------------------------------------

function myGetCorrespTVert iMeshVert iFace CurMesh =
(
	MeshFace = getFace CurMesh iFace
	TVFace = getTVFace CurMesh iFace
	if MeshFace.x == iMeshVert then
		return TVFace.x
	else if MeshFace.y == iMeshVert then 
		return TVFace.y
	else 
		return TVFace.z
)

strFileName = getSaveFileName caption:"Save Orbiter Mesh File" types:"Orbiter Mesh File *.msh|*.msh"
if strFileName != undefined then 
(
	fsMeshFile = CreateFile strFileName -- this line was added by Joshua Robertson
	if fsMeshFile == undefined then return();

	-- if we use a stringstream to dump the data to the output window
	-- we generate a little spacer to make it more clearly where the actual file begins
	if isKindOf fsMeshFile StringStream then
		format "\n\n*.msh File start below\n\n---------X8-------------snip-----------------X8-------------------\n" to:fsMeshFile

	-- first write the header
	format "MSHX1\n" to:fsMeshFile

	iNumObjects = 0
	arrMeshes = #()
	arrNames = #()

	-- now copy all objects to an internal array and convert them to 
	-- Editable_Mesh for exporting
	for iObIndex = 1 to $objects.count do
	(
		AktObject = snapshot $objects[iObIndex]

		if AktObject != undefined then 
		(
			iNumObjects += 1
			arrMeshes[iNumObjects] = AktObject
			arrNames[iNumObjects] = $objects[iObIndex].name
		)
		else
			print "Skipping " + $objects[iObIndex].name + "! Not convertable to Editable_Mesh !"
	)

	-- output GROUPS tag with number of meshes
	format "GROUPS %\n" iNumObjects to:fsMeshFile

	arrTextures = #()

	-- assemble an array with all the texturenames used in the scene now
	-- so we can calculate the correct texture index for each mesh
	for CurMat in sceneMaterials do 
	(
		if CurMat.diffusemap != undefined then
		(
			strTextureFileExploded = filterString CurMat.diffusemap.filename "\\."
			strTextureFile = strTextureFileExploded[strTextureFileExploded.count - 1]
			if findItem arrTextures strTextureFile == 0 then append arrTextures strTextureFile
		)
	)

	for iAktMesh = 1 to arrMeshes.count do
	(
		CurMesh = arrMeshes[iAktMesh]

		print ("Generating geometry data for " + arrNames[iAktMesh] + " as Group " + (iAktMesh-1) as string)

		-- output material index for current mesh
		if CurMesh.Material == undefined then iMat = 0
		else iMat = findItem sceneMaterials CurMesh.Material
		format "MATERIAL %\n" iMat to:fsMeshFile

		-- output texture index 
		iTexIndex = 0
		if iMat > 0 and sceneMaterials[iMat].diffusemap != undefined then
		(
			strTextureFileExploded = filterString sceneMaterials[iMat].diffusemap.filename "\\."
			strTextureFile = strTextureFileExploded[strTextureFileExploded.count - 1]
			iTexIndex = findItem arrTextures strTextureFile 
		)
		format "TEXTURE %\n" iTexIndex to:fsMeshFile 

		-- determine whether texture coordinates have to be created 
		bGenTexCoords = false
		if CurMesh.material != undefined and CurMesh.material.diffuseMap != undefined then 
			bGenTexCoords = true

		-- the following part is for handling vertices with multiple normals 
		-- since orbiter can only handle 1 normal per vertice those with multiple
		-- normals will now be split up so that each of the resulting vertices 
		-- has only one normal vector

		-- first we need to store the texture vertices, because our texturemapping
		-- would get lost, if we added additional texturevertices to the mesh, so we 
		-- store it all here and add texturevertices to this array later
		arrTVerts = #()
		for iAktTVert = 1 to CurMesh.numtverts do
			append arrTVerts (getTVert CurMesh iAktTVert)

		
		struct SFaceGroup (iSmoothGroup, arrFaces, iTVert)
		for iCurVert = 1 to CurMesh.verts.count do
		(
			-- now we determine for every vertex of the mesh, how many normals 
			-- it has.. this is done by examining the smoothing groups of the
			-- attached faces
			arrFaceGroups = #()
			arrVertFaces = meshop.getFacesUsingVert CurMesh iCurVert
			for iAktFace = 1 to arrVertFaces.count do
			(
				if arrVertFaces[iAktFace] then 
				(
					-- first we store one FaceGroup for every face of the vertex
					if bGenTexCoords then
						global FaceGroup = SFaceGroup iSmoothGroup:(getFaceSmoothGroup CurMesh iAktFace) arrFaces:#(iAktFace) iTVert:(myGetCorrespTVert iCurVert iAktFace CurMesh)
					else 
						global FaceGroup = SFaceGroup iSmoothGroup:(getFaceSmoothGroup CurMesh iAktFace) arrFaces:#(iAktFace) iTVert:-1
					append arrFaceGroups FaceGroup
				)
			)
			if arrFaceGroups.count > 0 then
			(
				do
				(
					-- now we check wich groups can be put together into one group
					-- according to their smoothgroup setting
					-- this is done until no two groups share any smoothing groups 
					bChanged = false
					for iAktFaceGroup = 1 to arrFaceGroups.count do
					(
						iCompFaceGroup = iAktFaceGroup+1
						while iCompFaceGroup <= arrFaceGroups.count do
						(
							if ((bit.and arrFaceGroups[iAktFaceGroup].iSmoothGroup arrFaceGroups[iCompFaceGroup].iSmoothGroup != 0) and (arrFaceGroups[iAktFaceGroup].iTVert == arrFaceGroups[iCompFaceGroup].iTVert)) then
							(
								iNewSmoothGroup = bit.or arrFaceGroups[iAktFaceGroup].iSmoothGroup arrFaceGroups[iCompFaceGroup].iSmoothGroup
								join arrFaceGroups[iAktFaceGroup].arrFaces arrFaceGroups[iCompFaceGroup].arrFaces
								arrFaceGroups[iAktFaceGroup].iSmoothGroup = iNewSmoothGroup
								deleteItem arrFaceGroups iCompFaceGroup
								bChanged = true
							)
							else
								iCompFaceGroup += 1
						)
					)
				)
				while bChanged
				-- the resulting number of groups is the number of normal vectors
				-- for our vertex thus the total number of vertices needed at this 
				-- position

				if bGenTexCoords then
					arrTVerts[iCurVert] = getTVert CurMesh (arrFaceGroups[1].iTVert)

				-- now if there are more than 1 group, a new vertex is created
				-- for each additional group and the faces of the group are
				-- reconnected to the new vertex
				for iAktGroup = 2 to arrFaceGroups.count do
				(
					iNewVertIndex = setNumVerts CurMesh (CurMesh.numverts+1) true
					setVert CurMesh iNewVertIndex CurMesh.verts[iCurVert].pos
					-- add texture vertex as well 
					if bGenTexCoords then
						arrTVerts[iNewVertIndex] = getTVert CurMesh arrFaceGroups[iAktGroup].iTVert
					for iAktFace in arrFaceGroups[iAktGroup].arrFaces do
					(
						AktFace = getFace CurMesh iAktFace
						if AktFace.x == iCurVert then 
							AktFace.x = iNewVertIndex
						else if AktFace.y == iCurVert then 
							AktFace.y = iNewVertIndex
						else 
							AktFace.z = iNewVertIndex
						setFace CurMesh iAktFace AktFace
					)								
				)
			)
			else
				print ("Ignoring isolated vertex " + iCurVert as string)
		)
		update CurMesh
		-- now that we got this done we can generate .msh data for the current mesh
				
		-- start of geometry data 
		format "GEOM % % ; %\n" CurMesh.numverts CurMesh.numfaces arrNames[iAktMesh] to:fsMeshFile

		for iCurVert = 1 to CurMesh.numverts do
		(
			-- get vertex
			vertex = CurMesh.verts[iCurVert]

			-- get normal vector
			pntVNormal = getNormal CurMesh iCurVert
			if bGenTexCoords then
			(
				-- get texture coordinates				
				pntTVert = arrTVerts[iCurVert]

				-- output it all to the stream and perform right hand to left hand conversion
				format "% % % % % % % %\n" (vertex.pos.x*-1.0) vertex.pos.y vertex.pos.z (pntVNormal.x*-1.0) pntVNormal.y pntVNormal.z pntTVert.x (pntTVert.y*-1.0+1.0) to:fsMeshFile
			)
			else  -- version without texture coordinates
				format "% % % % % %\n" (vertex.pos.x*-1.0) vertex.pos.y vertex.pos.z (pntVNormal.x*-1.0) pntVNormal.y pntVNormal.z to:fsMeshFile
		)

		-- now output the face definitions 
		for iCurFace = 1 to CurMesh.numfaces do
		(
			face = getFace CurMesh iCurFace
			format "% % %\n" (face.x as integer - 1) (face.z as integer - 1) (face.y as integer - 1) to:fsMeshFile
		)
	)

	-- after all geometry data is generated, output the material definitions
	print "Generating Material definitions"
	format "MATERIALS %\n" sceneMaterials.count to:fsMeshFile
	for CurMat in sceneMaterials do
		format "%\n" CurMat.name to:fsMeshFile
	for CurMat in sceneMaterials do
	(
		format "MATERIAL %\n" CurMat.name to:fsMeshFile

		-- output the materials color definitions, opacity is stored in all color definitions
		format "% % % %\n" (CurMat.diffuse.r / 255.0) (CurMat.diffuse.g / 255.0) (CurMat.diffuse.b / 255.0) (CurMat.opacity / 100.0) to:fsMeshFile
		format "% % % %\n" (CurMat.ambient.r / 255.0) (CurMat.ambient.g / 255.0) (CurMat.ambient.b / 255.0) (CurMat.opacity / 100.0) to:fsMeshFile
		format "% % % % %\n" (CurMat.specular.r / 255.0) (CurMat.specular.g / 255.0) (CurMat.specular.b / 255.0) (CurMat.opacity / 100.0) (CurMat.specularlevel) to:fsMeshFile
		format "% % % %\n" (CurMat.selfIllumColor.r / 255.0) (CurMat.selfIllumColor.g / 255.0) (CurMat.selfIllumColor.b / 255.0) (CurMat.opacity / 100.0) to:fsMeshFile
	)
	-- output texture names with a .dds extension, textures have to be converted to .dds format manually
	print "Generating Texture List"
	format "TEXTURES %\n" arrTextures.count to:fsMeshFile
	for strCurTex in arrTextures do
		format "%.dds\n" strCurTex to:fsMeshFile

	-- delete all the Editable_Mesh copies of the scene objects from the scene
	for CurMesh in arrMeshes do
		delete CurMesh

	-- if we use a stringstream to dump the data to the output window
	-- we generate a little stuff here for marking the end of the .msh file
	if isKindOf fsMeshFile StringStream then
		format "---------X8--------snip-----------------X8-------------------\nEnd of *.msh File\n" to:fsMeshFile


	print "Done generating data."
		
	close fsMeshFile
)
